import logging

from choicelib import choice_in_order
from typing_extensions import Protocol


class JSONModule(Protocol):
    def loads(self, s: str) -> dict:
        ...

    def dumps(self, o: dict) -> str:
        ...

    def load(self, f: str) -> dict:
        ...

    def dump(self, o: dict, f: str) -> None:
        ...


json: JSONModule = choice_in_order(
    ["ujson", "hyperjson", "orjson"], do_import=True, default="json"
)
logger = choice_in_order(["loguru"], do_import=True, default="logging")

if logger.__name__ == "logging":
    logger = logging.getLogger("vkbottle")
elif logger.__name__ == "loguru":
    logger = getattr(logger, "logger")



from typing import TYPE_CHECKING, Optional

from pydantic import root_validator

from ..base import BaseMessageMin

if TYPE_CHECKING:

    from vkbottle.api import ABCAPI


class MessageMin(BaseMessageMin):
    user_id: Optional[int] = None

    @property
    def is_mentioned(self) -> bool:
        return False if not self.mention else self.mention.id == self.user_id

    @root_validator(pre=True)
    def __foreign_messages(cls, values):
        foreign_messages = []
        if values.get("fwd_messages"):
            foreign_messages.extend(values["fwd_messages"])
        if values.get("reply_message"):
            foreign_messages.append(values["reply_message"])
        for foreign_message in foreign_messages:
            foreign_message["unprepared_ctx_api"] = values["unprepared_ctx_api"]
            foreign_message["replace_mention"] = values["replace_mention"]
            foreign_message["user_id"] = values["user_id"]
        return values


MessageMin.update_forward_refs()


async def message_min(
    message_id: int, ctx_api: "ABCAPI", replace_mention: bool = True
) -> "MessageMin":
    message_object = (await ctx_api.request("messages.getById", {"message_ids": message_id}))[
        "response"
    ]["items"][0]
    return MessageMin(
        **message_object, unprepared_ctx_api=ctx_api, replace_mention=replace_mention
    )



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Type, Union

if TYPE_CHECKING:
    from vkbottle.dispatch.rules import ABCRule
    from vkbottle.dispatch.views import ABCView
    from vkbottle.dispatch.views.abc import ABCMessageView, ABCRawEventView
    from vkbottle.tools.dev.mini_types.base.message import BaseMessageMin

    LabeledMessageHandler = Callable[..., Callable[["BaseMessageMin"], Any]]
    LabeledHandler = Callable[..., Callable[[Any], Any]]


class ABCLabeler(ABC):
    message_view: "ABCMessageView"
    raw_event_view: "ABCRawEventView"
    custom_rules: Dict[str, Type["ABCRule"]]
    auto_rules: List["ABCRule"]
    raw_event_auto_rules: List["ABCRule"]
    rule_config: Dict[str, Any]

    @abstractmethod
    def message(self, *rules: "ABCRule", **custom_rules) -> "LabeledMessageHandler":
        pass

    @abstractmethod
    def chat_message(self, *rules: "ABCRule", **custom_rules) -> "LabeledMessageHandler":
        pass

    @abstractmethod
    def private_message(self, *rules: "ABCRule", **custom_rules) -> "LabeledMessageHandler":
        pass

    @abstractmethod
    def raw_event(
        self,
        event: Union[str, List[str]],
        dataclass: Any,
        *rules: "ABCRule",
        **custom_rules,
    ) -> "LabeledHandler":
        pass

    @abstractmethod
    def views(self) -> Dict[str, "ABCView"]:
        pass

    @abstractmethod
    def load(self, labeler: Any):
        pass



import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

import vbml

from vkbottle.dispatch.handlers import FromFuncHandler
from vkbottle.dispatch.rules import ABCRule, base
from vkbottle.dispatch.views.abc import ABCRawEventView

from .abc import ABCLabeler

if TYPE_CHECKING:
    from vkbottle.dispatch.views import ABCView
    from vkbottle.dispatch.views.abc import ABCMessageView

    from .abc import LabeledMessageHandler


DEFAULT_CUSTOM_RULES: Dict[str, Type["ABCRule"]] = {
    "from_chat": base.PeerRule,
    "mention": base.MentionRule,
    "command": base.CommandRule,
    "from_user": base.FromUserRule,
    "peer_ids": base.FromPeerRule,
    "sticker": base.StickerRule,
    "attachment": base.AttachmentTypeRule,
    "levenshtein": base.LevenshteinRule,
    "lev": base.LevenshteinRule,
    "length": base.MessageLengthRule,
    "action": base.ChatActionRule,
    "payload": base.PayloadRule,
    "payload_contains": base.PayloadContainsRule,
    "payload_map": base.PayloadMapRule,
    "func": base.FuncRule,
    "coro": base.CoroutineRule,
    "coroutine": base.CoroutineRule,
    "state": base.StateRule,
    "state_group": base.StateGroupRule,
    "regexp": base.RegexRule,
    "regex": base.RegexRule,
    "macro": base.MacroRule,
    "text": base.VBMLRule,
}


class BaseLabeler(ABCLabeler):
    def __init__(
        self,
        message_view: "ABCMessageView",
        raw_event_view: "ABCRawEventView",
        custom_rules: Optional[Dict[str, Type["ABCRule"]]] = None,
        auto_rules: Optional[List["ABCRule"]] = None,
        raw_event_auto_rules: Optional[List["ABCRule"]] = None,
    ):
        self.message_view = message_view
        self.raw_event_view = raw_event_view

        self.custom_rules = custom_rules or DEFAULT_CUSTOM_RULES
        self.auto_rules = auto_rules or []
        self.raw_event_auto_rules = raw_event_auto_rules or []

        # Rule config is accessible from every single custom rule
        self.rule_config: Dict[str, Any] = {
            "vbml_flags": re.DOTALL,  # Flags for VBMLRule
            "vbml_patcher": vbml.Patcher(),  # Patcher for VBMLRule
        }

    @property
    def vbml_ignore_case(self) -> bool:
        """Gets ignore case flag from rule config flags"""
        return re.IGNORECASE in self.rule_config["flags"]

    @vbml_ignore_case.setter
    def vbml_ignore_case(self, ignore_case: bool):
        """Adds ignore case flag to rule config flags or removes it"""
        if not ignore_case:
            self.rule_config["vbml_flags"] ^= re.IGNORECASE
        else:
            self.rule_config["vbml_flags"] |= re.IGNORECASE

    @property
    def vbml_patcher(self) -> vbml.Patcher:
        return self.rule_config["vbml_patcher"]

    @vbml_patcher.setter
    def vbml_patcher(self, patcher: vbml.Patcher):
        self.rule_config["vbml_patcher"] = patcher

    @property
    def vbml_flags(self) -> re.RegexFlag:
        return self.rule_config["vbml_flags"]

    @vbml_flags.setter
    def vbml_flags(self, flags: re.RegexFlag):
        self.rule_config["vbml_flags"] = flags

    def message(
        self, *rules: "ABCRule", blocking: bool = True, **custom_rules
    ) -> "LabeledMessageHandler":
        assert all(isinstance(rule, ABCRule) for rule in rules), (
            "All rules must be subclasses of ABCRule or rule shortcuts "
            "(https://vkbottle.readthedocs.io/ru/latest/high-level/routing/rules/)"
        )

        def decorator(func):
            self.message_view.handlers.append(
                FromFuncHandler(
                    func,
                    *rules,
                    *self.auto_rules,
                    *self.get_custom_rules(custom_rules),
                    blocking=blocking,
                )
            )
            return func

        return decorator

    def chat_message(
        self, *rules: "ABCRule", blocking: bool = True, **custom_rules
    ) -> "LabeledMessageHandler":
        assert all(isinstance(rule, ABCRule) for rule in rules), (
            "All rules must be subclasses of ABCRule or rule shortcuts "
            "(https://vkbottle.readthedocs.io/ru/latest/high-level/routing/rules/)"
        )

        def decorator(func):
            self.message_view.handlers.append(
                FromFuncHandler(
                    func,
                    base.PeerRule(),
                    *rules,
                    *self.auto_rules,
                    *self.get_custom_rules(custom_rules),
                    blocking=blocking,
                )
            )
            return func

        return decorator

    def private_message(
        self, *rules: "ABCRule", blocking: bool = True, **custom_rules
    ) -> "LabeledMessageHandler":
        assert all(isinstance(rule, ABCRule) for rule in rules), (
            "All rules must be subclasses of ABCRule or rule shortcuts "
            "(https://vkbottle.readthedocs.io/ru/latest/high-level/routing/rules/)"
        )

        def decorator(func):
            self.message_view.handlers.append(
                FromFuncHandler(
                    func,
                    base.PeerRule(False),
                    *rules,
                    *self.auto_rules,
                    *self.get_custom_rules(custom_rules),
                    blocking=blocking,
                )
            )
            return func

        return decorator

    def load(self, labeler: "BaseLabeler"):
        self.message_view.handlers.extend(labeler.message_view.handlers)
        self.message_view.middlewares.extend(labeler.message_view.middlewares)
        self.raw_event_view.middlewares.extend(labeler.raw_event_view.middlewares)
        for event, handler_basements in labeler.raw_event_view.handlers.items():
            event_handlers = self.raw_event_view.handlers.setdefault(event, [])
            event_handlers.extend(handler_basements)

    def get_custom_rules(self, custom_rules: Dict[str, Any]) -> List["ABCRule"]:
        return [self.custom_rules[k].with_config(self.rule_config)(v) for k, v in custom_rules.items()]  # type: ignore

    def views(self) -> Dict[str, "ABCView"]:
        return {"message": self.message_view, "raw": self.raw_event_view}



from .auth import AuthError, UserAuth
from .ctx_tool import BaseContext
from .delayed_task import DelayedTask
from .event_data import OpenAppEvent, OpenLinkEvent, ShowSnackbarEvent
from .keyboard import (
    EMPTY_KEYBOARD,
    ABCAction,
    Callback,
    Keyboard,
    KeyboardButtonColor,
    Location,
    OpenLink,
    Text,
    VKApps,
    VKPay,
)
from .loop_wrapper import LoopWrapper
from .mini_types import BotTypes, UserTypes
from .storage import ABCStorage, CtxStorage
from .template import TemplateElement, template_gen
from .uploader import (
    AudioUploader,
    BaseUploader,
    DocMessagesUploader,
    DocUploader,
    DocWallUploader,
    GraffitiUploader,
    PhotoChatFaviconUploader,
    PhotoFaviconUploader,
    PhotoMarketUploader,
    PhotoMessageUploader,
    PhotoToAlbumUploader,
    PhotoUploader,
    PhotoWallUploader,
    VideoUploader,
    VoiceMessageUploader,
)
from .utils import load_blueprints_from_package, run_in_task, run_sync
from .vkscript_converter import vkscript

__all__ = (
    "ABCAction",
    "ABCStorage",
    "AudioUploader",
    "AudioUploader",
    "AuthError",
    "BaseContext",
    "BaseUploader",
    "BaseUploader",
    "BotTypes",
    "Callback",
    "CtxStorage",
    "DelayedTask",
    "DocMessagesUploader",
    "DocMessagesUploader",
    "DocUploader",
    "DocUploader",
    "DocWallUploader",
    "DocWallUploader",
    "EMPTY_KEYBOARD",
    "GraffitiUploader",
    "GraffitiUploader",
    "Keyboard",
    "KeyboardButtonColor",
    "load_blueprints_from_package",
    "Location",
    "LoopWrapper",
    "OpenAppEvent",
    "OpenLink",
    "OpenLinkEvent",
    "PhotoChatFaviconUploader",
    "PhotoFaviconUploader",
    "PhotoMarketUploader",
    "PhotoMessageUploader",
    "PhotoToAlbumUploader",
    "PhotoUploader",
    "PhotoWallUploader",
    "run_in_task",
    "run_sync",
    "ShowSnackbarEvent",
    "template_gen",
    "TemplateElement",
    "Text",
    "UserAuth",
    "UserTypes",
    "VideoUploader",
    "VideoUploader",
    "VKApps",
    "VKPay",
    "vkscript",
    "VoiceMessageUploader",
    "VoiceMessageUploader",
)



from typing import TYPE_CHECKING, Any, NoReturn, Union

from vkbottle.modules import json, logger

from .abc import ABCResponseValidator

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API


class JSONResponseValidator(ABCResponseValidator):
    """Default response json-parse validator
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/response-validator.md
    """

    async def validate(
        self,
        method: str,
        data: dict,
        response: Any,
        ctx_api: Union["ABCAPI", "API"],
    ) -> Union[Any, NoReturn]:
        if isinstance(response, dict):
            return response
        elif isinstance(response, str):
            return json.loads(response)

        logger.info(
            f"VK returned object of invalid type ({type(response)})."
            f"Request will be rescheduled with {ctx_api.request_rescheduler.__class__.__name__!r}"
        )

        return await self.validate(
            method,
            data,
            await ctx_api.request_rescheduler.reschedule(ctx_api, method, data, response),
            ctx_api,
        )



from enum import Enum
from typing import Any

from pydantic import BaseModel, validator


class BaseStateGroup(str, Enum):
    pass


def get_state_repr(state: BaseStateGroup) -> str:
    return f"{state.__class__.__name__}:{state.value}"


class StatePeer(BaseModel):
    peer_id: int
    state: str
    payload: dict = {}

    @validator("state", pre=True)
    def validate_state(cls, v: Any) -> str:
        if isinstance(v, BaseStateGroup):
            return get_state_repr(v)
        elif isinstance(v, str):
            return v
        raise ValueError(f"State value must be `string` or `BaseStateGroup`, got `{type(v)}`")



from typing import TYPE_CHECKING, Dict, List, NamedTuple, Type

from vkbottle_types.events import UserEventType

from vkbottle.dispatch.return_manager.user import UserMessageReturnHandler
from vkbottle.dispatch.views.abc.raw import ABCRawEventView

if TYPE_CHECKING:
    from vkbottle_types.events import BaseUserEvent

    from vkbottle.dispatch.handlers import ABCHandler


class UserHandlerBasement(NamedTuple):
    dataclass: Type["BaseUserEvent"]
    handler: "ABCHandler"


class RawUserEventView(ABCRawEventView[list]):
    handlers: Dict[UserEventType, List["UserHandlerBasement"]]

    def __init__(self):
        super().__init__()
        self.handlers = {}
        self.handler_return_manager = UserMessageReturnHandler()

    def get_handler_basements(self, event: list) -> List["UserHandlerBasement"]:
        return self.handlers[UserEventType(self.get_event_type(event))]

    def get_event_model(
        self, handler_basement: "UserHandlerBasement", event: list
    ) -> "BaseUserEvent":
        return handler_basement.dataclass(*event)

    @staticmethod
    def get_event_type(event: list) -> int:
        return event[0]

    async def process_event(self, event: list) -> bool:
        try:
            event_type = UserEventType(self.get_event_type(event))
        except ValueError:
            event_type = UserEventType.UNDEFINED_EVENT
        return event_type in self.handlers



from abc import ABC, abstractmethod
from asyncio import AbstractEventLoop
from typing import TYPE_CHECKING, NoReturn

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI
    from vkbottle.polling import ABCPolling


class ABCFramework(ABC):
    api: "ABCAPI"

    @property
    @abstractmethod
    def polling(self) -> "ABCPolling":
        pass

    @property
    @abstractmethod
    def loop(self) -> AbstractEventLoop:
        pass

    @abstractmethod
    async def run_polling(self) -> NoReturn:  # type: ignore
        pass

    @abstractmethod
    def run_forever(self) -> NoReturn:  # type: ignore
        pass



import itertools
from typing import Iterable

from .abc import ABCTokenGenerator


class ConsistentTokenGenerator(ABCTokenGenerator):
    def __init__(self, tokens: Iterable[str]):
        self.tokens = itertools.cycle(tokens)

    async def get_token(self) -> str:
        return next(self.tokens)



import ast
import random
import string
from typing import Callable, Iterable

from typing_extensions import Protocol

from .base_converter import Converter, ConverterError

CALL_REPLACEMENTS = {
    "append": "push",
    "pop": "pop",
}
CALL_STRING = ["join", "strip", "split"]

converter = Converter()
find = converter.find_definition


def dispatch_keywords(keywords: Iterable, assigner: str = ":", sep: str = ","):
    return sep.join(f"{param.arg}{assigner}{find(param.value)}" for param in keywords)


def random_string(length: int) -> str:
    return "".join(random.choice(string.ascii_lowercase) for _ in range(length))


def to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    return components[0].lower() + "".join(x.title() for x in components[1:])


@converter(ast.Assign)
def assign(d: ast.Assign):
    left = d.targets
    left_ = [find(target) for target in left if target.__class__ == ast.Name]
    right = find(d.value)
    return "var " + ",".join(f"{target}={right}" for target in left_) + ";"


@converter(ast.Add)
@converter(ast.UAdd)
def add_operator(_: ast.Add):
    return "+"


@converter(ast.Sub)
@converter(ast.USub)
def sub_operator(_: ast.Sub):
    return "-"


@converter(ast.Mult)
def mult_operator(_: ast.Mult):
    return "*"


@converter(ast.Div)
def div_operator(_: ast.Div):
    return "/"


@converter(ast.Pow)
def pow_operator(_: ast.Pow):
    return "**"


@converter(ast.RShift)
def rshift_operator(_: ast.RShift):
    return ">>"


@converter(ast.LShift)
def lshift_operator(_: ast.LShift):
    return "<<"


@converter(ast.BitAnd)
def bitand_operator(_: ast.BitAnd):
    return "&"


@converter(ast.BitOr)
def bitor_operator(_: ast.BitOr):
    return "|"


@converter(ast.Mod)
def mod_operator(_: ast.Mod):
    return "%"


@converter(ast.Gt)
def gt_operator(_: ast.Gt):
    return ">"


@converter(ast.Lt)
def lt_operator(_: ast.Lt):
    return "<"


@converter(ast.GtE)
def gt_operator_1(_: ast.GtE):
    return ">="


@converter(ast.LtE)
def gt_operator_2(_: ast.LtE):
    return "<="


@converter(ast.Eq)
def gt_operator_3(_: ast.Eq):
    return "=="


@converter(ast.NotEq)
def gt_operator_4(_: ast.Gt):
    return "!="


@converter(ast.And)
def and_operator(_: ast.And):
    return "&&"


@converter(ast.Or)
def or_operator(_: ast.Or):
    return "||"


@converter(ast.AugAssign)
def aug_assign(d: ast.AugAssign):
    operator = find(d.op)
    value = find(d.value)
    target = find(d.target)
    return f"{target} = {target} {operator} {value};"


@converter(ast.Constant)
def constant(d: ast.Constant):
    return d.value


@converter(ast.Name)
def name(d: ast.Name):
    return d.id


@converter(ast.While)
def while_cycle(d: ast.While):
    if d.orelse:
        raise ConverterError("You can't use while or/else in vkscript")
    body = "".join(find(line) for line in d.body)
    return "while(" + find(d.test) + "){" + body + "};"


@converter(ast.For)
def for_cycle(d: ast.For):
    random_iter_name = f"__iter_{random_string(5)}__"
    body = "".join(find(line) for line in d.body)
    return (
        f"var {random_iter_name} = {find(d.iter)};"
        f"while({random_iter_name}.length > 0){{var {find(d.target)}={random_iter_name}.pop();{body}}};"
    )


@converter(ast.If)
def if_statement(d: ast.If):
    return (
        "if("
        + find(d.test)
        + "){"
        + "".join(find(li) for li in d.body)
        + "}else{"
        + ("".join(find(e) for e in d.orelse) if len(d.orelse) else "")
        + "};"
    )


@converter(ast.Call)
def call(d: ast.Call):
    func: ast.Attribute = d.func  # type: ignore
    calls = []

    while isinstance(func, ast.Attribute):
        calls.append(func.attr)
        func = func.value  # type: ignore
    if func.__class__ == ast.Str:
        if calls[0] in CALL_STRING:
            return str(find(d.args[0])) + "." + calls[0] + "(" + find(func) + ")"
        elif calls[0] == "format":
            raise ConverterError("Use f-strings instead of str.format")
        raise ConverterError("String formatter")

    if func.id.lower() == "api":
        params = dispatch_keywords(d.keywords)
        return "API." + ".".join(map(to_camel_case, calls[::-1])) + "({" + params + "})"
    elif func.id == "len":
        return f"{find(d.args[0])}.length"
    elif calls and calls[0] in CALL_REPLACEMENTS:
        args = ",".join(find(arg) for arg in d.args)
        return find(d.func.value) + "." + CALL_REPLACEMENTS[calls[0]] + "(" + args + ")"  # type: ignore
    elif calls[0] in CALL_STRING:
        return find(func) + "." + calls[0] + "(" + find(d.args[0]) + ")"
    raise ConverterError(f"Call for {getattr(d.func, 'attr', d.func.__dict__)} is not referenced")


@converter(ast.Pass)
def pass_expr(_: ast.Pass):
    return ""


@converter(ast.Expr)
def expr(d: ast.Expr):
    return find(d.value) + ";"


@converter(ast.Module)
def module(d: ast.Module):
    return find(d.body)


@converter(ast.BinOp)
def bin_operation(d: ast.BinOp):
    return f"{find(d.left)}{find(d.op)}{find(d.right)}"


@converter(ast.Compare)
def compare(d: ast.Compare):
    left = find(d.left)
    operations = [
        f"{left}{find(operator)}{find(comparator)}"
        for operator, comparator in zip(d.ops, d.comparators)
    ]
    return "&&".join(operations)


@converter(ast.BoolOp)
def bool_op(d: ast.BoolOp):
    return find(d.op).join(find(value) for value in d.values)


@converter(ast.UnaryOp)
def unary_op(d: ast.UnaryOp):
    return f"{find(d.op)}{find(d.operand)}"


@converter(ast.Subscript)
def subscript(d: ast.Subscript):
    value = find(d.value)
    if d.slice.__class__ == ast.Index:
        if d.slice.value.__class__ == str:  # type: ignore
            return f"{value}.{d.slice.value.s}"  # type: ignore
        return f"{value}[{find(d.slice.value)}]"  # type: ignore
    raise ConverterError(f"Slice {d.slice} is not referenced")


@converter(ast.Attribute)
def attribute(d: ast.Attribute):
    return f"{find(d.value)}.{d.attr}"


@converter(ast.Return)
def return_statement(d: ast.Return):
    value = "null" if d is None else find(d.value)
    return f"return {value};"


@converter(ast.Delete)
def delete_statement(d: ast.Delete):
    return "".join(f"delete {find(target)};" for target in d.targets)


@converter(ast.Dict)
def dict_type(d: ast.Dict):
    dict_s = ",".join(f"{find(key)}:{find(value)}" for key, value in zip(d.keys, d.values))
    return "{" + dict_s + "}"


@converter(ast.Num)
def num_type(d: ast.Num):
    return str(d.n)


@converter(ast.Str)
def str_type(d: ast.Num):
    return repr(d.s)


@converter(ast.JoinedStr)
def joined_str(d: ast.JoinedStr):
    return "+".join(find(value) for value in d.values)


@converter(ast.FormattedValue)
def formatted_value(d: ast.FormattedValue):
    return find(d.value)


@converter(ast.List)
@converter(ast.Tuple)
def list_type(d: ast.List):
    return "[" + ",".join(find(a) for a in d.elts) + "]"


@converter(ast.NameConstant)
def name_constant_type(d: ast.NameConstant):
    consts = {True: "true", False: "false", None: "null"}
    return consts[d.value]


class VKScriptFunction(Protocol):
    def __call__(self, **kwargs) -> str:
        ...


def vkscript(func: Callable) -> VKScriptFunction:
    def decorator(**context):
        return converter.scriptify(func, **context)

    return decorator



from typing import List, Optional

from pydantic import root_validator
from vkbottle_types.objects import ClientInfoForBots

from ..base.foreign_message import BaseForeignMessageMin


class ForeignMessageMin(BaseForeignMessageMin):
    group_id: Optional[int] = None
    client_info: Optional["ClientInfoForBots"] = None
    reply_message: Optional["ForeignMessageMin"] = None
    fwd_messages: Optional[List["ForeignMessageMin"]] = []

    @root_validator
    def __foreign_messages(cls, values):
        foreign_messages = []
        if values.get("fwd_messages"):
            foreign_messages.extend(values["fwd_messages"])
        if values.get("reply_message"):
            foreign_messages.append(values["reply_message"])
        for foreign_message in foreign_messages:
            foreign_message.unprepared_ctx_api = values["unprepared_ctx_api"]
            foreign_message.replace_mention = values["replace_mention"]
            foreign_message.group_id = values["group_id"]
            foreign_message.client_info = values["client_info"]
        return values

    @property
    def is_mentioned(self) -> bool:
        if not (self.mention and self.group_id):
            return False
        return self.mention.id == -self.group_id


ForeignMessageMin.update_forward_refs()



from typing import TYPE_CHECKING, Any, Dict, NoReturn, Type, Union

from vkbottle.exception_factory import CaptchaError, VKAPIError
from vkbottle.modules import logger

from .abc import ABCResponseValidator

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI, API


SPECIFIC_ERRORS: Dict[int, Type[VKAPIError]] = {14: CaptchaError}


class VKAPIErrorResponseValidator(ABCResponseValidator):
    """Default vk api error response validator
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/response-validator.md
    """

    async def validate(
        self,
        method: str,
        data: dict,
        response: Any,
        ctx_api: Union["ABCAPI", "API"],
    ) -> Union[Any, NoReturn]:
        if "error" not in response:
            if "response" not in response:
                # invalid response, just igrnore it
                return response
            elif isinstance(response["response"], list):
                errors = [item["error"] for item in response["response"] if "error" in item]
                if errors:
                    logger.debug(
                        f"{len(errors)} API error(s) in response wasn't handled: {errors}"
                    )
            return response

        if ctx_api.ignore_errors:
            return None
        error = response["error"]
        code = error.pop("error_code")
        exception = SPECIFIC_ERRORS.get(code, VKAPIError[code])

        if exception == CaptchaError and ctx_api.captcha_handler:
            key = await ctx_api.captcha_handler(exception(**error))  # type: ignore
            return await ctx_api.request(
                method, {**data, "captcha_sid": error["captcha_sid"], "captcha_key": key}
            )

        raise exception(**error)



from .abc import ABCAPI
from .api import API
from .request_rescheduler import ABCRequestRescheduler, BlockingRequestRescheduler
from .request_validator import DEFAULT_REQUEST_VALIDATORS, ABCRequestValidator
from .response_validator import DEFAULT_RESPONSE_VALIDATORS, ABCResponseValidator
from .token_generator import (
    ABCTokenGenerator,
    ConsistentTokenGenerator,
    SingleTokenGenerator,
    Token,
    get_token_generator,
)

__all__ = (
    "ABCAPI",
    "ABCRequestRescheduler",
    "ABCRequestValidator",
    "ABCResponseValidator",
    "ABCTokenGenerator",
    "API",
    "BlockingRequestRescheduler",
    "ConsistentTokenGenerator",
    "DEFAULT_REQUEST_VALIDATORS",
    "DEFAULT_RESPONSE_VALIDATORS",
    "get_token_generator",
    "SingleTokenGenerator",
    "Token",
)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI
    from vkbottle.exception_factory import ABCErrorHandler


class ABCCallback(ABC):
    """
    Abstract Callback class
    """

    @abstractmethod
    def get_secret_key(self) -> str:
        pass

    @abstractmethod
    async def setup_group_id(self) -> None:
        pass

    @abstractmethod
    async def find_server_id(self) -> Optional[int]:
        pass

    @abstractmethod
    async def add_callback_server(self) -> Any:
        pass

    @abstractmethod
    async def delete_callback_server(self, server_id: int) -> Any:
        pass

    @abstractmethod
    async def edit_callback_server(self, server_id: int, secret_key: Optional[str] = None) -> Any:
        pass

    @abstractmethod
    async def get_callback_confirmation_code(self) -> Any:
        pass

    @abstractmethod
    async def get_callback_servers(self, server_ids: Optional[List[int]] = None) -> Any:
        pass

    @abstractmethod
    async def get_callback_settings(self, server_id: int) -> Any:
        pass

    @abstractmethod
    async def set_callback_settings(
        self, server_id: int, params: Optional[Dict[str, bool]] = None
    ) -> Any:
        pass

    @property
    @abstractmethod
    def api(self) -> "ABCAPI":
        pass

    @api.setter
    def api(self, new_api: "ABCAPI"):
        pass

    @abstractmethod
    def construct(
        self, api: "ABCAPI", error_handler: Optional["ABCErrorHandler"] = None
    ) -> "ABCCallback":
        pass



import asyncio
from typing import TYPE_CHECKING, Any, Optional

from aiohttp import ClientSession, TCPConnector

from vkbottle.modules import json as json_module

from .abc import ABCHTTPClient

if TYPE_CHECKING:
    from aiohttp import ClientResponse


class AiohttpClient(ABCHTTPClient):
    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        session: Optional[ClientSession] = None,
        json_processing_module: Optional[Any] = None,
        optimize: bool = False,
        **kwargs,
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.json_processing_module = json_processing_module or json_module

        if optimize:
            kwargs["skip_auto_headers"] = {"User-Agent"}
            kwargs["raise_for_status"] = True

        self.session = session

        self._session_params = kwargs

    async def request_raw(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> "ClientResponse":
        if not self.session:
            self.session = ClientSession(
                connector=TCPConnector(ssl=False),
                json_serialize=self.json_processing_module.dumps,
                **self._session_params,
            )
        async with self.session.request(url=url, method=method, data=data, **kwargs) as response:
            await response.read()
            return response

    async def request_json(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> dict:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.json(
            encoding="utf-8", loads=self.json_processing_module.loads, content_type=None
        )

    async def request_text(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> str:
        response = await self.request_raw(url, method, data, **kwargs)
        return await response.text(encoding="utf-8")

    async def request_content(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> bytes:
        response = await self.request_raw(url, method, data, **kwargs)
        return response._body

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

    def __del__(self):
        if self.session and not self.session.closed:
            if self.session._connector is not None and self.session._connector_owner:
                self.session._connector.close()
            self.session._connector = None


class SingleAiohttpClient(AiohttpClient):
    __instance__ = None

    def __new__(cls, *args, **kwargs):
        if cls.__instance__ is not None:
            return cls.__instance__
        return super().__new__(cls)

    def __aexit__(self, exc_type, exc_val, exc_tb):
        pass  # no need to close session in this case



from .abc import ABCTokenGenerator, Token
from .consistent import ConsistentTokenGenerator
from .single import SingleTokenGenerator
from .util import get_token_generator

__all__ = (
    "ABCTokenGenerator",
    "ConsistentTokenGenerator",
    "get_token_generator",
    "SingleTokenGenerator",
    "Token",
)



import inspect
import re
import types
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    Iterable,
    List,
    Optional,
    Pattern,
    Tuple,
    Type,
    Union,
)

import vbml

from vkbottle.dispatch.dispenser import get_state_repr
from vkbottle.tools.dev.mini_types.base import BaseMessageMin
from vkbottle.tools.validator import (
    ABCValidator,
    CallableValidator,
    EqualsValidator,
    IsInstanceValidator,
)

if TYPE_CHECKING:
    from vkbottle.dispatch.dispenser.base import BaseStateGroup

from .abc import ABCRule

__all__ = (
    "PeerRule",
    "MentionRule",
    "CommandRule",
    "VBMLRule",
    "StickerRule",
    "FromPeerRule",
    "AttachmentTypeRule",
    "LevenshteinRule",
    "MessageLengthRule",
    "ChatActionRule",
    "PayloadRule",
    "PayloadContainsRule",
    "PayloadMapRule",
    "FromUserRule",
    "FuncRule",
    "CoroutineRule",
    "StateRule",
    "StateGroupRule",
    "RegexRule",
    "MacroRule",
)


DEFAULT_PREFIXES = ["!", "/"]
PayloadMap = List[Tuple[str, Union[type, Callable[[Any], bool], ABCValidator, Any]]]
PayloadMapStrict = List[Tuple[str, ABCValidator]]
PayloadMapDict = Dict[str, Union[dict, type]]


class PeerRule(ABCRule[BaseMessageMin]):
    def __init__(self, from_chat: bool = True):
        self.from_chat = from_chat

    async def check(self, event: BaseMessageMin) -> bool:
        return self.from_chat is (event.peer_id != event.from_id)


class MentionRule(ABCRule[BaseMessageMin]):
    def __init__(self, mention_only: bool = True):
        self.mention_only = mention_only

    async def check(self, event: BaseMessageMin) -> bool:
        if self.mention_only:
            return event.is_mentioned and not event.text
        else:
            return event.is_mentioned


class CommandRule(ABCRule[BaseMessageMin]):
    def __init__(
        self,
        command_text: Union[str, Tuple[str, int]],
        prefixes: Optional[List[str]] = None,
        args_count: int = 0,
        sep: str = " ",
    ):
        self.prefixes = prefixes or DEFAULT_PREFIXES
        self.command_text = command_text if isinstance(command_text, str) else command_text[0]
        self.args_count = args_count if isinstance(command_text, str) else command_text[1]
        self.sep = sep

    async def check(self, event: BaseMessageMin) -> Union[dict, bool]:
        for prefix in self.prefixes:
            if self.args_count == 0 and event.text == prefix + self.command_text:
                return True
            if self.args_count > 0 and event.text.startswith(prefix + self.command_text + " "):
                args = event.text[len(prefix + self.command_text) + 1 :].split(self.sep)
                if len(args) != self.args_count:
                    return False
                elif any(len(arg) == 0 for arg in args):
                    return False
                return {"args": tuple(args)}
        return False


class VBMLRule(ABCRule[BaseMessageMin]):
    def __init__(
        self,
        pattern: Union[str, "vbml.Pattern", Iterable[Union[str, "vbml.Pattern"]]],
        patcher: Optional["vbml.Patcher"] = None,
        flags: Optional[re.RegexFlag] = None,
    ):
        flags = flags or self.config.get("vbml_flags") or re.DOTALL

        if isinstance(pattern, str):
            pattern = [vbml.Pattern(pattern, flags=flags or self.config.get("vbml_flags"))]
        elif isinstance(pattern, vbml.Pattern):
            pattern = [pattern]
        elif isinstance(pattern, Iterable):
            pattern = [
                p if isinstance(p, vbml.Pattern) else vbml.Pattern(p, flags=flags) for p in pattern
            ]

        self.patterns = pattern
        self.patcher = patcher or self.config.get("vbml_patcher") or vbml.Patcher()

    async def check(self, event: BaseMessageMin) -> Union[Optional[dict], bool]:
        for pattern in self.patterns:
            result = self.patcher.check(pattern, event.text)
            if result not in (None, False):
                return result
        return False


class RegexRule(ABCRule[BaseMessageMin]):
    def __init__(self, regexp: Union[str, List[str], Pattern, List[Pattern]]):
        if isinstance(regexp, Pattern):
            regexp = [regexp]
        elif isinstance(regexp, str):
            regexp = [re.compile(regexp)]
        elif isinstance(regexp, list):
            regexp = [re.compile(exp) for exp in regexp]

        self.regexp = regexp

    async def check(self, event: BaseMessageMin) -> Union[dict, bool]:
        for regexp in self.regexp:
            match = re.match(regexp, event.text)
            if match:
                return {"match": match.groups()}
        return False


class StickerRule(ABCRule[BaseMessageMin]):
    def __init__(self, sticker_ids: Union[List[int], int] = None):
        sticker_ids = sticker_ids or []
        if isinstance(sticker_ids, int):
            sticker_ids = [sticker_ids]
        self.sticker_ids = sticker_ids

    async def check(self, event: BaseMessageMin) -> bool:
        if not event.attachments:
            return False
        elif not event.attachments[0].sticker:
            return False
        else:
            if not self.sticker_ids:
                if event.attachments[0].sticker.sticker_id:
                    return True
            elif event.attachments[0].sticker.sticker_id in self.sticker_ids:
                return True
        return False


class FromPeerRule(ABCRule[BaseMessageMin]):
    def __init__(self, peer_ids: Union[List[int], int]):
        if isinstance(peer_ids, int):
            peer_ids = [peer_ids]
        self.peer_ids = peer_ids

    async def check(self, event: BaseMessageMin) -> bool:
        return event.peer_id in self.peer_ids


class AttachmentTypeRule(ABCRule[BaseMessageMin]):
    def __init__(self, attachment_types: Union[List[str], str]):
        if not isinstance(attachment_types, list):
            attachment_types = [attachment_types]
        self.attachment_types = attachment_types

    async def check(self, event: BaseMessageMin) -> bool:
        if not event.attachments:
            return False
        return all(
            attachment.type.value in self.attachment_types for attachment in event.attachments
        )


class ForwardMessagesRule(ABCRule[BaseMessageMin]):
    async def check(self, event: BaseMessageMin) -> bool:
        return bool(event.fwd_messages)


class ReplyMessageRule(ABCRule[BaseMessageMin]):
    async def check(self, event: BaseMessageMin) -> bool:
        return bool(event.reply_message)


class GeoRule(ABCRule[BaseMessageMin]):
    async def check(self, event: BaseMessageMin) -> bool:
        return bool(event.geo)


class LevenshteinRule(ABCRule[BaseMessageMin]):
    def __init__(self, levenshtein_texts: Union[List[str], str], max_distance: int = 1):
        if isinstance(levenshtein_texts, str):
            levenshtein_texts = [levenshtein_texts]
        self.levenshtein_texts = levenshtein_texts
        self.max_distance = max_distance

    @staticmethod
    def distance(a: str, b: str) -> int:
        n, m = len(a), len(b)
        if n > m:
            a, b = b, a
            n, m = m, n

        current_row = range(n + 1)
        for i in range(1, m + 1):
            previous_row, current_row = current_row, [i] + [0] * n  # type: ignore
            for j in range(1, n + 1):
                add, delete, change = (
                    previous_row[j] + 1,
                    current_row[j - 1] + 1,
                    previous_row[j - 1],
                )
                if a[j - 1] != b[i - 1]:
                    change += 1
                current_row[j] = min(add, delete, change)  # type: ignore

        return current_row[n]

    async def check(self, event: BaseMessageMin) -> bool:
        return any(
            self.distance(event.text, levenshtein_text) <= self.max_distance
            for levenshtein_text in self.levenshtein_texts
        )


class MessageLengthRule(ABCRule[BaseMessageMin]):
    def __init__(self, min_length: int):
        self.min_length = min_length

    async def check(self, event: BaseMessageMin) -> bool:
        return len(event.text) >= self.min_length


class ChatActionRule(ABCRule[BaseMessageMin]):
    def __init__(self, chat_action_types: Union[List[str], str]):
        if isinstance(chat_action_types, str):
            chat_action_types = [chat_action_types]
        self.chat_action_types = chat_action_types

    async def check(self, event: BaseMessageMin) -> bool:
        if not event.action:
            return False
        elif event.action.type.value in self.chat_action_types:
            return True
        return False


class PayloadRule(ABCRule[BaseMessageMin]):
    def __init__(self, payload: Union[dict, List[dict]]):
        if isinstance(payload, dict):
            payload = [payload]
        self.payload = payload

    async def check(self, event: BaseMessageMin) -> bool:
        return event.get_payload_json() in self.payload


class PayloadContainsRule(ABCRule[BaseMessageMin]):
    def __init__(self, payload_particular_part: dict):
        self.payload_particular_part = payload_particular_part

    async def check(self, event: BaseMessageMin) -> bool:
        if event.payload is None:
            return False
        payload = event.get_payload_json()
        if not isinstance(payload, dict):
            return False
        return all(payload.get(k) == v for k, v in self.payload_particular_part.items())


class PayloadMapRule(ABCRule[BaseMessageMin]):
    def __init__(self, payload_map: Union[PayloadMap, PayloadMapDict]):
        if isinstance(payload_map, dict):
            payload_map = self.transform_to_map(payload_map)
        self.payload_map = self.transform_to_callbacks(payload_map)

    @classmethod
    def transform_to_map(cls, payload_map_dict: PayloadMapDict) -> PayloadMap:
        """Transforms PayloadMapDict to PayloadMap"""
        payload_map = []
        for (k, v) in payload_map_dict.items():
            if isinstance(v, dict):
                v = cls.transform_to_map(v)  # type: ignore
            payload_map.append((k, v))
        return payload_map  # type: ignore

    @classmethod
    def transform_to_callbacks(cls, payload_map: PayloadMap) -> PayloadMapStrict:
        """Transforms PayloadMap to PayloadMapStrict"""
        for i, (key, value) in enumerate(payload_map):
            if isinstance(value, type):
                value = IsInstanceValidator(value)
            elif isinstance(value, list):
                value = cls.transform_to_callbacks(value)
            elif isinstance(value, types.FunctionType):
                value = CallableValidator(value)
            elif not isinstance(value, ABCValidator):
                value = EqualsValidator(value)
            payload_map[i] = (key, value)
        return payload_map  # type: ignore

    @classmethod
    async def match(cls, payload: dict, payload_map: PayloadMapStrict):
        """Matches payload with payload_map recursively"""
        for (k, validator) in payload_map:
            if k not in payload:
                return False
            elif isinstance(validator, list):
                if not isinstance(payload[k], dict):
                    return False
                elif not await cls.match(payload[k], validator):
                    return False
            elif not await validator.check(payload[k]):
                return False
        return True

    async def check(self, event: BaseMessageMin) -> Union[dict, bool]:
        payload = event.get_payload_json()
        if not isinstance(payload, dict):
            return False
        if await self.match(payload, self.payload_map):
            return payload
        return False


class FromUserRule(ABCRule[BaseMessageMin]):
    def __init__(self, from_user: bool = True):
        self.from_user = from_user

    async def check(self, event: BaseMessageMin) -> bool:
        return self.from_user is (event.from_id > 0)


class FuncRule(ABCRule[BaseMessageMin]):
    def __init__(self, func: Callable[[BaseMessageMin], Union[bool, Awaitable]]):
        self.func = func

    async def check(self, event: BaseMessageMin) -> Union[dict, bool]:
        if inspect.iscoroutinefunction(self.func):
            return await self.func(event)  # type: ignore
        return self.func(event)  # type: ignore


class CoroutineRule(ABCRule[BaseMessageMin]):
    def __init__(self, coroutine: Coroutine):
        self.coro = coroutine

    async def check(self, event: BaseMessageMin) -> Union[dict, bool]:
        return await self.coro


class StateRule(ABCRule[BaseMessageMin]):
    def __init__(
        self,
        state: Optional[Union[List["BaseStateGroup"], "BaseStateGroup"]] = None,
    ):
        if not isinstance(state, list):
            state = [] if state is None else [state]
        self.state = [get_state_repr(s) for s in state]

    async def check(self, event: BaseMessageMin) -> bool:
        if event.state_peer is None:
            return not self.state
        return event.state_peer.state in self.state


class StateGroupRule(ABCRule[BaseMessageMin]):
    def __init__(
        self,
        state_group: Optional[Union[List[Type["BaseStateGroup"]], Type["BaseStateGroup"]]] = None,
    ):
        if not isinstance(state_group, list):
            state_group = [] if state_group is None else [state_group]
        self.state_group = [group.__name__ for group in state_group]

    async def check(self, event: BaseMessageMin) -> bool:
        if event.state_peer is None:
            return not self.state_group
        group_name = event.state_peer.state.split(":", maxsplit=1)[0]
        return group_name in self.state_group


try:
    import macro  # type: ignore
except ImportError:
    macro = None


class MacroRule(ABCRule[BaseMessageMin]):
    def __init__(self, pattern: Union[str, List[str]]):
        if macro is None:
            raise RuntimeError("macro must be installed to use MacroRule")

        if isinstance(pattern, str):
            pattern = [pattern]
        self.patterns = list(map(macro.Pattern, pattern))

    async def check(self, event: BaseMessageMin) -> Union[dict, bool]:
        for pattern in self.patterns:
            result = pattern.check(event.text)
            if result is not None:
                return result
        return False



import traceback
from inspect import iscoroutinefunction
from typing import Any, Callable, Optional, Tuple, Type, Union

from vkbottle.modules import logger


def swear(
    exception: Union[
        BaseException, Type[BaseException], Tuple[Union[BaseException, Type[BaseException]], ...]
    ],
    exception_handler: Optional[Callable] = None,
    just_log: bool = False,
    just_return: bool = False,
) -> Any:
    """Swear catcher allows to handle exceptions | Used as a decorator
    :param exception: Exception(s) to handle
    :param exception_handler: async exception handler
    :param just_log: should swear handler log the error
    :param just_return: should the exception just be returned
    >>> @swear(RuntimeError, just_return=True)
    >>> def function():
    >>>     raise RuntimeError("Oh no!")
    >>> function()
    >>> RuntimeError("Oh no!")
    """

    if not isinstance(exception, tuple):
        exception = (exception,)

    def decorator(func: Callable):
        async def asynchronous_wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except exception as e:
                if exception_handler is not None:
                    return await exception_handler(e, *args, **kwargs)
                elif just_log:
                    logger.error(
                        f"{func.__name__} (handling with swear) has thrown an exception: \n\n{traceback.format_exc()}"
                    )
                elif just_return:
                    return e
            finally:
                logger.debug(f"Function {func.__name__} was handled with swear")

        def synchronous_wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exception as e:
                if exception_handler is not None:
                    return exception_handler(e, *args, **kwargs)
                elif just_log:
                    logger.error(
                        f"{func.__name__} (handling with swear) has thrown an exception: \n\n{traceback.format_exc()}"
                    )
                elif just_return:
                    return e
            finally:
                logger.debug(f"Function {func.__name__} was handled with swear")

        if iscoroutinefunction(func) and iscoroutinefunction(exception_handler):
            return asynchronous_wrapper
        return synchronous_wrapper

    return decorator



from asyncio import get_event_loop
from typing import TYPE_CHECKING, NoReturn, Optional, Type, Union

from vkbottle.api import API
from vkbottle.dispatch import BuiltinStateDispenser, Router
from vkbottle.exception_factory import ErrorHandler
from vkbottle.framework.abc import ABCFramework
from vkbottle.framework.labeler import UserLabeler
from vkbottle.modules import logger
from vkbottle.polling import UserPolling
from vkbottle.tools import LoopWrapper, UserAuth

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop

    from vkbottle.api import ABCAPI, Token
    from vkbottle.dispatch import ABCRouter, ABCStateDispenser
    from vkbottle.exception_factory import ABCErrorHandler
    from vkbottle.framework.labeler import ABCLabeler
    from vkbottle.polling import ABCPolling


class User(ABCFramework):
    def __init__(
        self,
        token: Optional["Token"] = None,
        api: Optional["ABCAPI"] = None,
        polling: Optional["ABCPolling"] = None,
        loop: Optional["AbstractEventLoop"] = None,
        loop_wrapper: Optional[LoopWrapper] = None,
        router: Optional["ABCRouter"] = None,
        labeler: Optional["ABCLabeler"] = None,
        state_dispenser: Optional["ABCStateDispenser"] = None,
        error_handler: Optional["ABCErrorHandler"] = None,
        task_each_event: bool = True,
    ):
        self.api: Union["ABCAPI", API] = API(token) if token is not None else api  # type: ignore
        self.error_handler = error_handler or ErrorHandler()
        self.loop_wrapper = loop_wrapper or LoopWrapper()
        self.labeler = labeler or UserLabeler()
        self.state_dispenser = state_dispenser or BuiltinStateDispenser()
        self._polling = polling or UserPolling(self.api)
        self._router = router or Router()
        self._loop = loop
        self.task_each_event = task_each_event

    @property
    def polling(self) -> "ABCPolling":
        return self._polling.construct(self.api, self.error_handler)

    @property
    def router(self) -> "ABCRouter":
        return self._router.construct(
            views=self.labeler.views(),
            state_dispenser=self.state_dispenser,
            error_handler=self.error_handler,
        )

    @router.setter
    def router(self, new_router: "ABCRouter"):
        self._router = new_router

    @property
    def on(self) -> "ABCLabeler":
        return self.labeler

    @classmethod
    def direct_auth_sync(
        cls: Type["User"],
        login: str,
        password: str,
        client_id: int = None,
        client_secret: str = None,
        **kwargs,
    ):
        loop = get_event_loop()
        assert not loop.is_running(), "Event loop is already running, use direct_auth instead"
        return loop.run_until_complete(
            cls.direct_auth(
                login=login,
                password=password,
                client_id=client_id,
                client_secret=client_secret,
                **kwargs,
            )
        )

    @classmethod
    async def direct_auth(
        cls: Type["User"],
        login: str,
        password: str,
        client_id: int = None,
        client_secret: str = None,
        **kwargs,
    ):
        token = await UserAuth(client_id, client_secret).get_token(login, password)
        return cls(token=token, **kwargs)

    async def run_polling(self, custom_polling: Optional["ABCPolling"] = None) -> NoReturn:  # type: ignore
        polling = custom_polling or self.polling
        logger.info(f"Starting polling for {polling.api!r}")

        async for event in polling.listen():  # type: ignore
            logger.debug(f"New event was received: {event}")
            for update in event.get("updates", []):
                if not self.task_each_event:
                    await self.router.route(update, polling.api)
                else:
                    self.loop.create_task(self.router.route(update, polling.api))

    def run_forever(self) -> NoReturn:
        logger.info("Loop will be ran forever")
        self.loop_wrapper.add_task(self.run_polling())
        self.loop_wrapper.run_forever(self.loop)

    @property
    def loop(self) -> "AbstractEventLoop":
        if self._loop is None:
            self._loop = get_event_loop()
        return self._loop

    @loop.setter
    def loop(self, new_loop: "AbstractEventLoop"):
        self._loop = new_loop



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Optional, Type

from .abc import ABCErrorHandler

if TYPE_CHECKING:
    from .abc import AsyncFunc
from vkbottle.modules import logger


class ErrorHandler(ABCErrorHandler):
    def __init__(self, redirect_arguments: bool = False, raise_exceptions: bool = False):
        self.redirect_arguments = redirect_arguments
        self.raise_exceptions = raise_exceptions
        self.error_handlers = {}
        self.undefined_error_handler = None

    def register_error_handler(
        self, *error_types: Type[BaseException]
    ) -> Callable[["AsyncFunc"], "AsyncFunc"]:
        def decorator(handler: "AsyncFunc") -> "AsyncFunc":
            for error_type in error_types:
                self.error_handlers[error_type] = handler
            return handler

        return decorator

    def register_undefined_error_handler(self, handler: "AsyncFunc") -> "AsyncFunc":
        self.undefined_error_handler = handler
        return handler

    def lookup_handler(self, for_type: Type[BaseException]) -> Optional["AsyncFunc"]:
        for error_type in self.error_handlers:
            if issubclass(for_type, error_type):
                return self.error_handlers[error_type]

    async def handle(self, error: BaseException, *args, **kwargs) -> Any:
        handler = self.lookup_handler(type(error)) or self.undefined_error_handler

        if not handler:
            if self.raise_exceptions:
                raise error
            logger.exception(error)
            return

        if self.redirect_arguments:
            return await handler(error, *args, **kwargs)
        return await handler(error)

    def catch(self, func: "AsyncFunc") -> "AsyncFunc":
        @wraps(func)
        async def wrapper(*args, **kwargs) -> Any:
            try:
                await func(*args, **kwargs)
            except BaseException as error:
                return await self.handle(error, *args, **kwargs)

        return wrapper



from typing import TYPE_CHECKING, Any, Dict, Optional, Type

if TYPE_CHECKING:
    from .action import ABCAction
    from .color import KeyboardButtonColor


class KeyboardButton:
    def __init__(
        self,
        action: "ABCAction",
        color: Optional["KeyboardButtonColor"] = None,
        data: Optional[dict] = None,
    ):
        self.action = action
        self.color = color
        self.data = data

    @classmethod
    def from_typed(
        cls: Type["KeyboardButton"],
        action: "ABCAction",
        color: Optional["KeyboardButtonColor"] = None,
    ) -> "KeyboardButton":
        return cls(action, color, None)

    @classmethod
    def from_dict(cls: Type["KeyboardButton"], data: dict) -> "KeyboardButton":
        color = data.get("color")
        keyboard_data = {"action": data}
        if color is not None:
            keyboard_data["action"].pop("color")
            keyboard_data["color"] = color
        return cls(cls.action, cls.color, keyboard_data)  # type: ignore

    def get_data(self) -> dict:
        if self.data is not None:
            return self.data

        data: Dict[str, Any] = {"action": self.action.get_data()}
        if (
            self.action.type
            in (
                "text",
                "callback",
            )
            and self.color
        ):
            data["color"] = self.color.value
        return data



from vkbottle.framework.labeler import UserLabeler

from .dispatch.rules import base as rules
from .dispatch.views import user as views
from .framework.user import User, UserBlueprint, run_multibot
from .tools.dev.mini_types.bot import MessageMin

Message = MessageMin
Blueprint = UserBlueprint

__all__ = (
    "User",
    "Blueprint",
    "UserLabeler",
    "Message",
    "run_multibot",
    "rules",
    "views",
)



from asyncio import sleep
from typing import TYPE_CHECKING, Any, Callable, Coroutine

if TYPE_CHECKING:
    Handler = Callable[..., Coroutine[Any, Any, Any]]


class DelayedTask:
    def __init__(self, seconds: int, handler: "Handler", do_break: bool = False):
        self.seconds = seconds
        self.handler = handler
        self.do_break = do_break

    async def __call__(self, *args, **kwargs):
        while True:
            await sleep(self.seconds)
            await self.handler(*args, **kwargs)
            if self.do_break:
                break



from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, Dict, Optional, Type

AsyncFunc = Callable[..., Awaitable[Any]]


class ABCErrorHandler(ABC):
    error_handlers: Dict[Type[BaseException], AsyncFunc]
    undefined_error_handler: Optional[AsyncFunc]

    @abstractmethod
    def register_error_handler(
        self, *error_types: Type[BaseException]
    ) -> Callable[[AsyncFunc], AsyncFunc]:
        pass

    @abstractmethod
    def register_undefined_error_handler(self, handler: AsyncFunc) -> AsyncFunc:
        pass

    @abstractmethod
    async def handle(self, error: BaseException) -> Any:
        pass

    @abstractmethod
    def catch(self, func: AsyncFunc) -> AsyncFunc:
        pass



from abc import ABC, abstractmethod
from typing import Generic, Iterable, Type, TypeVar

T_contra = TypeVar("T_contra", contravariant=True)


class ABCRule(ABC, Generic[T_contra]):
    config: dict = {}

    @classmethod
    def with_config(cls, config: dict) -> Type["ABCRule"]:
        cls.config = config
        return cls

    @abstractmethod
    async def check(self, event: T_contra):
        pass

    def __and__(self, other: "ABCRule[T_contra]") -> "AndRule[T_contra]":
        return AndRule(self, other)

    def __or__(self, other: "ABCRule[T_contra]") -> "OrRule[T_contra]":
        return OrRule(self, other)

    def __invert__(self) -> "NotRule[T_contra]":
        return NotRule(self)

    def __repr__(self):
        return f"<{self.__class__.__qualname__}>"


class AndRule(ABCRule[T_contra], Generic[T_contra]):
    def __init__(self, *rules: ABCRule[T_contra]):
        self._rules = rules

    async def check(self, event: T_contra):
        context = {}

        for rule in self.rules:
            check_response = await rule.check(event)
            if check_response is False:
                return False
            elif isinstance(check_response, dict):
                context.update(check_response)

        return context

    @property
    def rules(self) -> Iterable[ABCRule[T_contra]]:
        return self._rules


class NotRule(ABCRule[T_contra]):
    def __init__(self, *rules: ABCRule[T_contra]):
        self._rules = rules

    async def check(self, event: T_contra):
        for rule in self.rules:
            check_response = await rule.check(event)
            if check_response is False:
                return True
        return False

    @property
    def rules(self) -> Iterable[ABCRule[T_contra]]:
        return self._rules


class OrRule(ABCRule[T_contra]):
    def __init__(self, *rules: ABCRule[T_contra]):
        self._rules = rules

    async def check(self, event: T_contra):
        for rule in self.rules:
            check_response = await rule.check(event)
            if check_response is not False:
                return check_response
        return False

    @property
    def rules(self) -> Iterable[ABCRule[T_contra]]:
        return self._rules



from abc import ABC, abstractmethod
from typing import NoReturn, Union


class ABCRequestValidator(ABC):
    """Abstract Request Validator class
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/request_validator.md
    """

    @abstractmethod
    async def validate(self, request: dict) -> Union[dict, NoReturn]:
        pass



from .dev import (
    EMPTY_KEYBOARD,
    ABCAction,
    ABCStorage,
    AudioUploader,
    AuthError,
    BaseContext,
    BaseUploader,
    BotTypes,
    Callback,
    CtxStorage,
    DelayedTask,
    DocMessagesUploader,
    DocUploader,
    DocWallUploader,
    GraffitiUploader,
    Keyboard,
    KeyboardButtonColor,
    Location,
    LoopWrapper,
    OpenAppEvent,
    OpenLink,
    OpenLinkEvent,
    PhotoChatFaviconUploader,
    PhotoFaviconUploader,
    PhotoMarketUploader,
    PhotoMessageUploader,
    PhotoToAlbumUploader,
    PhotoUploader,
    PhotoWallUploader,
    ShowSnackbarEvent,
    TemplateElement,
    Text,
    UserAuth,
    UserTypes,
    VideoUploader,
    VKApps,
    VKPay,
    VoiceMessageUploader,
    load_blueprints_from_package,
    run_in_task,
    run_sync,
    template_gen,
    vkscript,
)
from .production import keyboard_gen
from .validator import ABCValidator, CallableValidator, EqualsValidator, IsInstanceValidator

__all__ = (
    "ABCAction",
    "ABCStorage",
    "ABCValidator",
    "AudioUploader",
    "AuthError",
    "BaseContext",
    "BaseUploader",
    "BotTypes",
    "CallableValidator",
    "Callback",
    "CtxStorage",
    "DelayedTask",
    "DocMessagesUploader",
    "DocUploader",
    "DocWallUploader",
    "EMPTY_KEYBOARD",
    "EqualsValidator",
    "GraffitiUploader",
    "IsInstanceValidator",
    "keyboard_gen",
    "Keyboard",
    "KeyboardButtonColor",
    "load_blueprints_from_package",
    "Location",
    "LoopWrapper",
    "OpenAppEvent",
    "OpenLink",
    "OpenLinkEvent",
    "PhotoChatFaviconUploader",
    "PhotoFaviconUploader",
    "PhotoMarketUploader",
    "PhotoMessageUploader",
    "PhotoToAlbumUploader",
    "PhotoUploader",
    "PhotoWallUploader",
    "run_in_task",
    "run_sync",
    "ShowSnackbarEvent",
    "template_gen",
    "TemplateElement",
    "Text",
    "UserAuth",
    "UserTypes",
    "VideoUploader",
    "VKApps",
    "VKPay",
    "vkscript",
    "VoiceMessageUploader",
)



from abc import abstractmethod
from typing import Generic, Optional, TypeVar

from .view import ABCView

T_contra = TypeVar("T_contra", list, dict, contravariant=True)
F_contra = TypeVar("F_contra", contravariant=True)


class ABCDispenseView(ABCView[T_contra], Generic[T_contra, F_contra]):
    @abstractmethod
    def get_state_key(self, event: F_contra) -> Optional[int]:
        pass



from typing import TYPE_CHECKING, List, Optional

from vkbottle.modules import json

from .button import KeyboardButton

if TYPE_CHECKING:
    from .action import ABCAction
    from .button import KeyboardButtonColor


class Keyboard:
    def __init__(self, one_time: bool = False, inline: bool = False):
        self.one_time = one_time
        self.inline = inline
        self.buttons: List[List[KeyboardButton]] = []

    def row(self) -> "Keyboard":
        if len(self.buttons) and not len(self.buttons[-1]):
            raise RuntimeError("Last row is empty!")
        self.buttons.append([])
        return self

    def add(
        self, action: "ABCAction", color: Optional["KeyboardButtonColor"] = None
    ) -> "Keyboard":
        if not len(self.buttons):
            self.row()
        button = KeyboardButton.from_typed(action, color)
        self.buttons[-1].append(button)
        return self

    def schema(self, rows: List[List[dict]]):
        for row in rows:
            self.row()
            for button in row:
                self.buttons[-1].append(KeyboardButton.from_dict(button))
        return self

    def get_json(self) -> str:
        data = json.dumps(
            {
                "one_time": self.one_time,
                "inline": self.inline,
                "buttons": [[button.get_data() for button in row] for row in self.buttons],
            }
        )

        if isinstance(data, bytes):
            return data.decode("utf-8")
        return data.encode("utf-8").decode("utf-8")

    def __str__(self) -> str:
        return self.get_json()



from pydantic import BaseModel

from vkbottle.modules import json

from .abc import ABCRequestValidator


class TranslateFriendlyTypesRequestValidator(ABCRequestValidator):
    async def validate(self, request: dict) -> dict:
        for k, v in request.copy().items():
            if isinstance(v, list):
                request[k] = ",".join(str(e) for e in v)
            elif isinstance(v, bool):
                request[k] = int(v)
            elif isinstance(v, BaseModel):
                request[k] = v.json(exclude_none=True, encoder=json.dumps)
            elif isinstance(v, dict):
                request[k] = json.dumps(await self.validate(v))
            elif v is None:
                del request[k]
        return request



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Awaitable, Callable, List, Optional

from vkbottle_types.categories import APICategories

if TYPE_CHECKING:
    from vkbottle.exception_factory import CaptchaError
    from vkbottle.http import ABCHTTPClient

    from .request_rescheduler import ABCRequestRescheduler
    from .request_validator import ABCRequestValidator
    from .response_validator import ABCResponseValidator
    from .token_generator import ABCTokenGenerator


class ABCAPI(APICategories, ABC):
    """Abstract API class
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/api/api.md
    """

    token_generator: "ABCTokenGenerator"
    ignore_errors: bool
    http_client: "ABCHTTPClient"
    request_rescheduler: "ABCRequestRescheduler"
    response_validators: List["ABCResponseValidator"]
    request_validators: List["ABCRequestValidator"]
    captcha_handler: Optional[Callable[["CaptchaError"], Awaitable]] = None

    @abstractmethod
    async def request(self, method: str, data: dict) -> dict:
        """Makes a single request opening a session"""

    @abstractmethod
    async def validate_response(self, method: str, data: dict, response: Any) -> Any:
        pass

    @abstractmethod
    async def validate_request(self, request: Any) -> Any:
        pass

    @property
    def api_instance(self) -> "ABCAPI":
        return self



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['vkbottle',
 'vkbottle.api',
 'vkbottle.api.request_rescheduler',
 'vkbottle.api.request_validator',
 'vkbottle.api.response_validator',
 'vkbottle.api.token_generator',
 'vkbottle.callback',
 'vkbottle.dispatch',
 'vkbottle.dispatch.dispenser',
 'vkbottle.dispatch.handlers',
 'vkbottle.dispatch.middlewares',
 'vkbottle.dispatch.return_manager',
 'vkbottle.dispatch.return_manager.bot',
 'vkbottle.dispatch.return_manager.user',
 'vkbottle.dispatch.rules',
 'vkbottle.dispatch.views',
 'vkbottle.dispatch.views.abc',
 'vkbottle.dispatch.views.bot',
 'vkbottle.dispatch.views.user',
 'vkbottle.exception_factory',
 'vkbottle.exception_factory.error_handler',
 'vkbottle.exception_factory.swear_handler',
 'vkbottle.framework',
 'vkbottle.framework.bot',
 'vkbottle.framework.labeler',
 'vkbottle.framework.user',
 'vkbottle.http',
 'vkbottle.polling',
 'vkbottle.tools',
 'vkbottle.tools.dev',
 'vkbottle.tools.dev.keyboard',
 'vkbottle.tools.dev.mini_types',
 'vkbottle.tools.dev.mini_types.base',
 'vkbottle.tools.dev.mini_types.bot',
 'vkbottle.tools.dev.mini_types.user',
 'vkbottle.tools.dev.storage',
 'vkbottle.tools.dev.template',
 'vkbottle.tools.dev.uploader',
 'vkbottle.tools.dev.vkscript_converter',
 'vkbottle.tools.production',
 'vkbottle.tools.production.legacies']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.8.0,<0.9.0',
 'aiohttp>=3.6.2,<4.0.0',
 'choicelib>=0.1.5,<0.2.0',
 'pydantic>=1.8.2,<2.0.0',
 'typing-extensions>=4.0.1,<5.0.0',
 'vbml>=1.1.post1,<2.0',
 'vkbottle-types>=5.131.134,<6.0.0',
 'watchgod>=0.7,<0.9']

setup_kwargs = {
    'name': 'vkbottle',
    'version': '4.3.0',
    'description': 'Homogenic! Customizable asynchronous VK API framework implementing comfort and speed',
    'long_description': '<p align="center">\n  <a href="https://github.com/vkbottle/vkbottle">\n    <img src="https://raw.githubusercontent.com/vkbottle/vkbottle/master/docs/logo.svg" width="175px" style="display: inline-block; border-radius: 5px">\n  </a>\n</p>\n<h1 align="center">\n  VKBottle 4.x\n</h1>\n<p align="center">\n  <img alt="GitHub Workflow Status" src="https://img.shields.io/github/workflow/status/vkbottle/vkbottle/CI?style=flat-square">\n  <img alt="GitHub repo size" src="https://img.shields.io/github/repo-size/vkbottle/vkbottle?style=flat-square">\n  <img alt="GitHub issues by-label" src="https://img.shields.io/github/issues/vkbottle/vkbottle/bug?style=flat-square">\n  <img alt="PyPI" src="https://img.shields.io/pypi/v/vkbottle?color=green&label=PyPI&style=flat-square">\n</p>\n\n> Кастомизируемый, быстрый и удобный фреймворк для работы с VK API\n\n## Документация\n\n[Туториал для новичков](https://vkbottle.readthedocs.io/ru/latest/tutorial/)\\\n[Техническая документация](https://vkbottle.readthedocs.io/ru/latest)\n\n## Установка\n\nУстановить новейшую версию можно командой:\n\n```shell\npip install -U https://github.com/vkbottle/vkbottle/archive/master.zip\n```\n\nУстановить последнюю версию VKBottle 4.x с PyPI можно командой:\n\n```shell\npip install vkbottle\n```\n\nЕсли вы ищете старые версии:\n- [`3.x`](https://github.com/vkbottle/vkbottle/tree/v3.0)\n- [`2.x`](https://github.com/vkbottle/vkbottle/tree/v2.0)\n\n## Hello World\n\n[Смотреть больше примеров!](https://github.com/vkbottle/vkbottle/tree/master/examples)\\\n[Почему VKBottle?](https://github.com/vkbottle/vkbottle/blob/master/docs/why_vkbottle.md)\n\n```python\nfrom vkbottle.bot import Bot\n\nbot = Bot("GroupToken")\n\n@bot.on.message()\nasync def handler(_) -> str:\n    return "Hello world!"\n\nbot.run_forever()\n```\n\n## Contributing\n\nПР поддерживаются! Перед созданием пулл реквеста ознакомьтесь с [CONTRIBUTION_GUIDE.md](CONTRIBUTION_GUIDE.md). Нам приятно видеть ваш вклад в развитие библиотеки. Задавайте вопросы в блоке Issues и в [**чате Telegram**](https://t.me/vkbottle_ru) / [**чате VK**](https://vk.me/join/AJQ1d7fBUBM_800lhEe_AwJj)!\n\n* Создатель [@timoniq](https://github.com/timoniq)\n* Мейнтейнер [@FeeeeK](https://github.com/FeeeeK)\n\n## Лицензия\n\nCopyright © 2019-2021 [timoniq](https://github.com/timoniq).\\\nCopyright © 2022 [FeeeeK](https://github.com/FeeeeK).\\\nЭтот проект имеет [MIT](https://github.com/vkbottle/vkbottle/blob/master/LICENSE) лицензию.\n',
    'author': 'timoniq',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vkbottle/vkbottle',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)



from typing import TYPE_CHECKING, Iterable, Type

from vkbottle.http import SingleAiohttpClient
from vkbottle.modules import logger
from vkbottle.polling import BotPolling

from .bot import Bot

if TYPE_CHECKING:
    from vkbottle.api import ABCAPI
    from vkbottle.polling import ABCPolling


def run_multibot(
    bot: Bot, apis: Iterable["ABCAPI"], polling_type: Type["ABCPolling"] = BotPolling
):
    """Add run_polling with polling constructed from derived apis
    :param bot: Bot main instance (api is not required)
    :param apis: Iterable of apis
    :param polling_type: polling type to be ran
    """
    for i, api_instance in enumerate(apis):
        logger.debug(f"Connecting API (index: {i})")
        polling = polling_type().construct(api_instance)
        api_instance.http_client = SingleAiohttpClient()
        bot.loop_wrapper.add_task(bot.run_polling(custom_polling=polling))
    bot.loop_wrapper.run_forever(bot.loop)



from vkbottle.modules import json

from .element import TemplateElement


def template_gen(*e: TemplateElement):
    """Generates dumped carousel out of box"""
    return json.dumps({"type": "carousel", "elements": [e.raw for e in e]})



from abc import ABC, abstractmethod
from typing import Any, Optional


class ABCHTTPClient(ABC):
    """Abstract class for http-clients
    Documentation: https://github.com/vkbottle/vkbottle/blob/master/docs/low-level/http/http-client.md
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        pass

    @abstractmethod
    async def request_raw(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> Any:
        pass

    @abstractmethod
    async def request_text(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> str:
        pass

    @abstractmethod
    async def request_json(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> dict:
        pass

    @abstractmethod
    async def request_content(
        self, url: str, method: str = "GET", data: Optional[dict] = None, **kwargs
    ) -> bytes:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass

    async def __aenter__(self) -> "ABCHTTPClient":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()
